/* VMS-specific implementation of mlock().
 * Adapted from the anonymous "mlock_vms.c" supplied by HP with its
 * GnuPG 1.2.3 kit.
 * 2006-06-22 SMS.
 *
 * Added support for fallback to protected shareable image to perform
 * locking.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>           /* Really the VMS-specific fake. */

#include <jpidef.h>
#include <prvdef.h>
#include <ssdef.h>
#include <starlet.h>
#include <stsdef.h>
#include <descrip.h>

static $DESCRIPTOR(protected_image,"GNUPG_PRIVILEGED_NONVAX");
static $DESCRIPTOR(lock_routine_name,"VMS_LOCK_MEMORY");
static $DESCRIPTOR(unlock_routine_name,"VMS_UNLOCK_MEMORY");

static int uwss_loaded = 0;
static int (*lock_mem_uwss) ( const void *, size_t length );
static int (*unlock_mem_uwss) ( void *, size_t length );

#include "config.h"
#include "util.h"

typedef struct _itemlist
{
        short buflen;
        short code;
        void *bufaddr;
        unsigned int *reslen;
} itemlist;

/*
 * Dynamically load protected shareable image that defines user written system
 * service that call lock a limited number of pages in memory, obviating the
 * need to install the program with syspriv.
 */
static int load_uwss ( )
{
    int status, lib$find_image_symbol( );
    int lib$sig_to_ret();
    uwss_loaded = 1;		/* at least we tried */
    lock_mem_uwss = 0;
    unlock_mem_uwss = 0;
    /*
     * Establish error handler to we don't get tracebacks on error.
     */
    vaxc$establish ( lib$sig_to_ret );

    status = lib$find_image_symbol ( &protected_image, 
	&lock_routine_name, &lock_mem_uwss, 0, 0 );
    if ( status & 1 ) {
	status = lib$find_image_symbol ( &protected_image,
		&unlock_routine_name, &unlock_mem_uwss, 0, 0 );
    } else {
	lock_mem_uwss = 0;
    }
    return status;
}


/* This function sets errno and vaxc$errno on error, and returns zero
 * (success) or non-zero errno (failure).
 */

int mlock( const void *addr, size_t len)
{
	int ret;
	int start_addr, end_addr;
	int lock_addr[2], retlock_addr[2];

	static int procpriv;
	itemlist ilist1[] = {
		{ (sizeof procpriv), JPI$_PROCPRIV, &procpriv, 0 },
		{ 0, 0, 0, 0 },
	};

	/*
	**  Check privileges.  (SYS$LCKPAG() requires PSWAPM.)
	*/

        /*
        ** Get process privileges.
        */
        ret = sys$getjpiw(0, 0, 0, ilist1, 0, 0, 0);

	if ((ret& STS$M_SEVERITY) != STS$K_SUCCESS)
        {
		/* GETJPIW failed. */
		errno = EVMSERR;
		vaxc$errno = ret;
		return errno;
        }

	/*
	** Test the process privileges for PSWAPM.
	*/
	if ((procpriv & PRV$M_PSWAPM) == 0)
	{
		/* We don't have PSWAPM. */

		if ( !uwss_loaded ) {
		    ret = load_uwss ( );
#ifdef TEST
		    printf ( "load_uwss status: %d\n", ret );
#endif
		}
		if ( lock_mem_uwss ) {
		    /* Use specified shareable image */
		    ret = lock_mem_uwss ( addr, len );
		    if ( ret & 1 ) return 0;   /* Return if memory locked OK */
		}

		log_warning(
		 "PSWAPM privilege is required for secure memory on VMS.\n");
		errno = EVMSERR;
		vaxc$errno = SS$_NOPRIV;
		return errno;
	}

	/*
	**  Set address range and then lock
	*/
	lock_addr[0] = (int)addr;		/* start address */
	lock_addr[1] = (int)addr+len-1;		/* end address */

	ret = sys$lckpag(lock_addr, retlock_addr, NULL);

#ifdef DEBUG
	printf("addr=%d, len=%d\n", (int)addr, len);
	printf("inaddr1=%x %d, inaddr2=%x %d\n", lock_addr[0], lock_addr[0], lock_addr[1], lock_addr[1]);
	printf("retaddr1=%x, retaddr2=%x\n", retlock_addr[0], retlock_addr[1]);
	printf("SYS$LCKPAG() >> %d\n", ret);
#endif
	if ((ret& STS$M_SEVERITY) == STS$K_SUCCESS)
	{
		/* Success. */
		return 0;
	}
	else
	{
		/* LCKPAG failed. */
		errno = EVMSERR;
		vaxc$errno = ret;
		return errno;
	}
}


/*-----------------------------------------------------------------------*/
#ifdef TEST
int main()
{
	int *a, *a_temp, ret, i, len = 100;

	a = malloc(sizeof(int)*len);
	a_temp = a;
	for(i=0; i<len; i++){
		*a_temp = 1;
		a_temp++;
	}

	ret = mlock(a, len);

	if(ret == 0)
		printf("mlock() for OpenVMS succeeded\n");
	else
		printf("mlock_vms() for OpenVMS failed\n");

	exit(0);
}
void g10_log_warning ( const char *ctlstr, ... )
{
    printf ( ctlstr );
}
#endif
