/****************************************************************************
 *									    *
 *									    *
 *   VMS Randomness-Gathering Code					    *
 *									    *
 *									    *
 ****************************************************************************/

/* This module is part of the cryptlib continuously seeded pseudorandom
   number generator.  For usage conditions, see lib_rand.c

   [Here is the notice from lib_rand.c:]

   This module and the misc/rnd*.c modules represent the cryptlib
   continuously seeded pseudorandom number generator (CSPRNG) as described in
   my 1998 Usenix Security Symposium paper "The generation of random numbers
   for cryptographic purposes".

   The CSPRNG code is copyright Peter Gutmann (and various others) 1996,
   1997, 1998, 1999, all rights reserved.  Redistribution of the CSPRNG
   modules and use in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice
      and this permission notice in its entirety.

   2. Redistributions in binary form must reproduce the copyright notice in
      the documentation and/or other materials provided with the distribution.

   3. A copy of any bugfixes or enhancements made must be provided to the
      author, <pgut001@cs.auckland.ac.nz> to allow them to be added to the
      baseline version of the code.

  ALTERNATIVELY, the code may be distributed under the terms of the GNU
  General Public License, version 2 or any later version published by the
  Free Software Foundation, in which case the provisions of the GNU GPL are
  required INSTEAD OF the above restrictions.

  Although not required under the terms of the GPL, it would still be nice if
  you could make any changes available to the author to allow a consistent
  code base to be maintained */
/*
 * Modification history:
 *    27-DEC-2007	Rewrite to use SYS$GETRMI function.   (DLJ)
 */

/* General includes */
#ifndef TEST_RMI
#include <config.h>
#endif

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#include <cmbdef.h>
#include <descrip.h>
#include <lib$routines.h>
#include <psldef.h>
#include <starlet.h>
#include <stsdef.h>
#include <rmidef.h>


#ifndef TEST_RMI
#include "types.h"  /* for byte and u32 typedefs */
#include "algorithms.h"
#include "util.h"
#endif
/*
 * Module-specific contants.
 */
#define RMI_RESULT_WORDS 16
#define RMI_DEFAULT_INTERVAL 990	/* 0.99 seconds */
#define DEBUG_ENV_VAR "GNUPG_RND_DEBUG"
#define DEBUG_AST_ENV_VAR "GNUPG_RNDVMS_RMI_DEBUG"
#define INTERVAL_ENV_VAR "GNUPG_RNDVMS_ENTROPY_INTERVAL"

static char *dbgfn;
static FILE *dbgfp = NULL;
/*
 * Define item list structure used by VMS system services as well a macro
 * to initialize.
 */
struct item_list {
    short int length, code;
    void *buffer;
    short int *retlen;
};
#define SET_ITEM(itm,c,l,b,r) itm.code=(c); itm.length=(l); itm.buffer=(b);\
   itm.retlen=(r)
/*
 * Define stucture to receive the output of the SYS$GETRMI call and make it
 * part of a union to facilitate manipulation of the result as raw words.
 */
static union {
    unsigned short uw[RMI_RESULT_WORDS];
    /*
     * To prevent alignment padding, order the data elements by decreasing data size.
     */
    struct {
	long long cpuidle, cpuint, cpukernel, cpuuser, tqetotal;
	long long systime;
	int lognam, mbreads, preads, opens, dd_hits;
	int dzroflts, exefaults, enqcvt, enqnew;
	int fcphit, fcpread, freflts, rdflts, access;
	unsigned short eol;		/* flag end of list */
    } rmi;
} entropy;
/*
 * Statically Define the item_list for the $GETRMI call, use helper macro.
 *
 * Note that some items are inter-related, reducing effective randomness of a 
 * sample.
 */
#define ENTROPY_ITEM(c,v) {sizeof(entropy.rmi.v), c, &entropy.rmi.v, 0}
static struct item_list  rmi_item[] = {
 ENTROPY_ITEM(RMI$_SYSTIME,systime),	/* system time */
 ENTROPY_ITEM(RMI$_CPUIDLE,cpuidle),	/* CPU idle ticks */
 ENTROPY_ITEM(RMI$_CPUINTSTK,cpuint),	/* interrupt stack ticks */
 ENTROPY_ITEM(RMI$_CPUKERNEL,cpukernel),/* kernel mode ticks */
 ENTROPY_ITEM(RMI$_CPUUSER,cpuuser),	/* user mode ticks */
 ENTROPY_ITEM(RMI$_TQETOTAL,tqetotal),	/* Cum. timer queue requests */
 ENTROPY_ITEM(RMI$_LOGNAM,lognam),	/* Logical name translations */
 ENTROPY_ITEM(RMI$_MBREADS,mbreads),	/* mailbox reads */
 ENTROPY_ITEM(RMI$_OPENS,opens),	/* file opens */
 ENTROPY_ITEM(RMI$_PREADS,preads),	/* page reads */
 ENTROPY_ITEM(RMI$_DIRDATA_HIT,dd_hits),/* dirdata cache hits */
 ENTROPY_ITEM(RMI$_DZROFLTS,dzroflts),	/* demand zero faults */
 ENTROPY_ITEM(RMI$_EXEFAULTS,exefaults),/* execute faults */
 ENTROPY_ITEM(RMI$_ENQCVT,enqcvt),      /* ENQ conversion requests */
 ENTROPY_ITEM(RMI$_ENQNEW,enqnew),	/* New enqueues */
 ENTROPY_ITEM(RMI$_FCPHIT,fcphit),	/* FCP cache hits */
 ENTROPY_ITEM(RMI$_FCPREAD,fcpread),	/* FCP read requests */
 ENTROPY_ITEM(RMI$_FREFLTS,freflts),	/* Free list faults */
 ENTROPY_ITEM(RMI$_RDFAULTS,rdflts),	/* read faults */
 ENTROPY_ITEM(RMI$_ACCESS,access),	/* Filename lookups */
 { 0, 0, &entropy.rmi.eol, 0 }		/* end of list. */
};
/*
 * After a scan, the entropy buffer is condensed to a rmi_scan_result struct.
 * Statistics of several scan results are tracked in a rmi_scan_quality struct.
 * An rmi_async_control struct holds context for the AST thread that does
 * new scans in the background at short intervals.
 */ 
struct rmi_scan_result {		/* 32 bytes of data */
    unsigned short uw[RMI_RESULT_WORDS];
};
struct rmi_scan_quality {
	int sample_count;	/* number of samples accumulated */
	int st_qual;		/* short-term quality */
	int lt_qual;		/* long term quality */
	int debug;		/* Print debug statements to dbgfp if true */
	/*
	 * Keep history of the number of times each bit in successively
         * collected scan results changed.  Divide bit_changes[i] by the
         * sample count to get the average number of times bit position i has
         * changed value.  Array bc_prev_4 and bc_prev_8 hold snapshots of the
         * bit_changes array at sample counts sc[0] and sc[1].
         */
	int sc[2];		/* Sample counts of bc_prev_* */
	int bit_changes[RMI_RESULT_WORDS*16];
	int bc_prev_4[RMI_RESULT_WORDS*16];
	int bc_prev_8[RMI_RESULT_WORDS*16];
};
struct rmi_async_control {
    short int scan_ef, stall_ef;
    int abort;
    int debug;
    int sample_num;			/* ID number of pending sample */
    int last_wait_num;			/* Last sample_num that completed wait*/
    struct rmi_scan_result *cur;	/* next result buffer to fill */
    struct rmi_scan_result *prev;	/* Last completed buffer */
    struct { int status, spare; } iosb;
    long long sample_interval;		/* VMS delta time */
    /*
     * Section for AST to comunicate with non_AST.  Consumer sets pending to
     * point to receive buffer and waits for staff_ef event flag.  Producer
     * fills receive buffer, pend_sample_num and pend_quality, then clears 
     * the pointer and sets flag. 
     */
    struct rmi_scan_result *pending;	/* Non-AST buffer */
    int pend_sample_num, pend_quality, pend_lt_quality;
    /*
     */
    struct rmi_scan_quality sq;
};
/*****************************************************************************/
/* Retrieve next RMI buffer read by AST thread or stall until next
 * read and have AST fill buffer.  Return value is indicator:
 *   2	2nd most recent sample returned.
 *   1  Most recent sample returned.
 *   0  Sampling thread was aborted due to error.
 *  -1  Thread stalled, next upcoming sample was returned.
 */
int rmi_await_next ( struct rmi_async_control *ctl, 
	struct rmi_scan_result *buf )
{
    int last_num, delta;
    struct rmi_scan_result *prev;
    /*
     * Syncrhonize with AST level and see if data available, preparing for
     * wait if not.
     */
    SYS$SETAST ( 0 );			/* disable AST execution */
    if ( ctl->pending ) {
	/* Bugcheck, already waiting for a buffer */
	delta = 0;
    } else {
	/*
	 * Return existing buffer if last pend_sample_num is less than
	 * most recent sample.
	 */
	delta = ctl->sample_num - ctl->pend_sample_num;
	if ( delta > 1 ) {
	    *buf = *ctl->prev;
	    ctl->pend_sample_num = ctl->sample_num - 1;
	    ctl->pend_lt_quality = ctl->sq.lt_qual;
	    ctl->pend_quality = ctl->sq.st_qual;
	    delta = 2;
	} else if ( delta > 0 ) {
	    *buf = *ctl->cur;
	    ctl->pend_sample_num = ctl->sample_num;
	    ctl->pend_lt_quality = ctl->sq.lt_qual;
	    ctl->pend_quality = ctl->sq.st_qual;;
	} else {
	   /* Stall required */
	   ctl->pending = buf;
	   delta = -1;
	}
    }
    SYS$SETAST ( 1 );
    if ( delta == 0 ) return 0;
    /*
     * Wait until pending pointer is clear.
     */
    while ( ctl->pending && !ctl->abort ) {
	SYS$WAITFR ( ctl->stall_ef );
	SYS$CLREF ( ctl->stall_ef );
    }

    return (ctl->abort) ? 0 : delta;
}
/*****************************************************************************/
/*
 * Condense the 52 words of data returned by $GETRMI call into a rmi_scan_result
 * structure. Return value is the intermediate number of folded words.
 */
static int condense_scan ( FILE *dbgf, struct rmi_scan_result *result )
{
    int i, j, folded_words;
    struct { int status; int s2; } iosb;
    unsigned short *rw, fw[RMI_RESULT_WORDS+8];
    /*
     * Fold each pair of words (longword) into a single word and eliminate any
     * zero results from the high half of the quadword items.
     */
    for ( i = 0, j = 0; &entropy.uw[i] < &entropy.rmi.eol; i+=2 ) {
	fw[j] = entropy.uw[i] ^ entropy.uw[i+1];
	if ( fw[j] ) j++;
	if ( j >= (RMI_RESULT_WORDS+8) ) break;
    }
    folded_words = j;
    /*
     * Reduce final result to 16 words (32 bytes) by processing folded words
     * exceeding that number.
     */
    rw = result->uw;
    memcpy ( rw, fw, sizeof(struct rmi_scan_result) );
    for ( i = RMI_RESULT_WORDS; i < folded_words; i++ ) {
	/*
	 * For each excess word, bit j being set will flip bit (i-16)+7 in
         * the result word j.
  	 */
	unsigned short mask, bit, toggle;
	mask = fw[i];
	bit = 1;
	toggle = 1<<(i-RMI_RESULT_WORDS+7);
	for ( j = 0; j < 16; j++ ) {
	    rw[j] ^= ((mask&bit)?toggle:0);
	    bit = bit*2;	/* shift left */
	}
    }
    /*
     * Show intermediate results.
     */
    if ( dbgf ) {
	fprintf ( dbgf, "Raw:\n" );
	for ( i = 0; i < folded_words*2; i++ ) {
	    fprintf ( dbgf, "%04x%c", entropy.uw[i], ((i&15)==15) ? '\n' : ' ' );
        }
	fprintf ( dbgf, "\nfolded: %d\n", folded_words );
	for ( i = 0; i < folded_words; i++ ) {
	    fprintf ( dbgf, "%04x%c", fw[i], (i&15)==15 ? '\n' : ' ' );
        }
        fprintf ( dbgf, "\nFinal:\n" );
    }
    return folded_words;
}
/*
 * Estimate a quality number between 0 and 256 from a distribution of bit 
 * changes for 8 sammples of 256 bits.  Distribution[2], for example, is the 
 * number of bit positions that had 2 value changes over the 8 samples. The 
 * ideal distribution is thus 1, 8, 28, 56, 70, 56, 28, 8, 1.  The function 
 * value returned represents the number of bits (out of 256) whose value is 
 * behaving 'randomly'.
 */
static int quality_estimate ( int *distribution, int min_est )
{
    int i, qual;
    static int expected[9] = { 1, 8, 28, 56, 70, 56, 28, 8, 1 };
#ifdef QUICK_QUAL
    /*
     * In the ideal dist., 3, 4, or 5 bits change 75% of the time.  Make our
     * our estimate summing these three values and re-normalzing.
     */
    qual = distribution[3] + distribution[4] + distribution[5];
    qual = (qual*4) / 3;
#else
    /*
     * Compute sum of square of differences from ideal distribution and use
     * square root of result as number of 'non-random' bits.
     */
    double chi_sq, observed, chi;
    chi_sq = 0.0;
    for ( i = 0; i < 9; i++ ) {
	observed = distribution[i];
	chi_sq += ((observed-expected[i])*(observed-expected[i])/expected[i]);
    }
    chi = sqrt(chi_sq);
    /*
     * Invert the result to get the number of 'random' bits.
     */
    if ( chi < 511.0 ) qual = 256 - (chi+0.5); else qual = -255;
#endif
    if ( qual > 256 ) qual = 256 + (256-qual);
    if ( qual < min_est ) qual = min_est;
    return qual;
    
}
/*******************************************************************************/
/* Calling rmi_sample makes a single sychronous call to SYS$GETRMI() and 
 * condenses the result to 32 bytes.  Return value is status of system service
 * call.
 */
int rmi_sample ( FILE *dbgf, struct rmi_scan_result *result )
{
    int status, i, j, folded_words;
    struct { int status; int s2; } iosb;
    unsigned short *rw, fw[RMI_RESULT_WORDS+8];
    /*
     * Load raw data into buffer, there is no 'W' form of SYS$GETRMI so
     * we must manually synch.
     */
    status = SYS$GETRMI ( 15, 0, 0, rmi_item, &iosb, 0, 0 );
    if ( status&1  ) status = SYS$SYNCH ( 15, &iosb );
    if ( status&1 ) status = iosb.status;
    if ( (status&1) == 1 ) condense_scan ( dbgf, result );

    return status;
}
/*
 * The collect quality function tracks rmi_scan_result buffers over
 * time and updates sq->st_qual and sq->lt_qual every 4th call.  Return value 
 * is raw number of bits that changes between current sample and previous.
 */
static int collect_quality ( struct rmi_scan_result *cur,
	struct rmi_scan_result *prev, struct rmi_scan_quality *sq )
{
    int bit_delta, i, j, n;
    unsigned short bit, mask;
    /*
     * Compare each bit in the current scan with the saved result and increment
     * the counters for each bit.
     */
    bit_delta = 0;		/* number of changed bits */
    for ( i = 0; i < RMI_RESULT_WORDS; i++ ) {
	if ( sq->debug ) fprintf ( dbgfp, "%04x%c", cur->uw[i], 
		(i==15) ? '\n' : ' ' );
	mask = cur->uw[i] ^ prev->uw[i];
	for ( j = 0, bit = 1; j < 16; j++ ) {
	    if ( bit & mask ) {
		bit_delta++;
		sq->bit_changes[(i*16)+j]++;
	    }
	    bit = bit*2;
	}
    }
    sq->sample_count++;
    /*
     * Save the change counts every 4 samples and re-compute quality
     */
    if ( (sq->sample_count % 4) == 0 ) {
 	int bin[9];
	/*
	 * Make a histogram to count the number of bits that
         * changed and use to revise quality estimate.
	 */
        memset ( bin, 0, sizeof(bin) );
        for ( i = 0; i < 256; i++ ) {
	    n = (sq->bit_changes[i]*8) / sq->sample_count;
	    bin[n]++;
        }
	sq->lt_qual = quality_estimate ( bin, 32 );
	if ( sq->debug ) fprintf ( dbgfp, 
		"change bins(s=%d): %d %d %d [%d %d %d] %d %d %d-> %d\n", 
		sq->sample_count, bin[0], bin[1],
		bin[2], bin[3], bin[4], bin[5], bin[6], bin[7],bin[8], sq->lt_qual );

	memset ( bin, 0, sizeof(bin) );
        for ( i = 0; i < 256; i++ ) {
	    n = (sq->bit_changes[i]-sq->bc_prev_8[i]);
	    if ( n > 8 ) n = 8;
	    bin[n]++;
        }
	sq->st_qual = quality_estimate ( bin, 32 );
	if ( sq->debug ) fprintf ( dbgfp, 
		"change bins(since %d): %d %d %d [%d %d %d] %d %d %d-> %d\n", 
		sq->sc[1], bin[0], bin[1],
		bin[2], bin[3], bin[4], bin[5], bin[6], bin[7], bin[8],
		sq->st_qual );
	/*
	 * Shift prev_4 changes array to prev_8 and put current
	 * in prev_4.
	 */
	sq->sc[1] = sq->sc[0];
	sq->sc[0] = sq->sample_count;
	memcpy ( sq->bc_prev_8, sq->bc_prev_4, sizeof(sq->bc_prev_4) );
	memcpy ( sq->bc_prev_4, sq->bit_changes, sizeof(sq->bit_changes) );
    } else if ( (sq->sample_count < 4) ) {
	/*
	 * Just start, make guess as short term quality.
	 */
	sq->st_qual = bit_delta;
	if ( bit_delta > (RMI_RESULT_WORDS*8) )
	    bit_delta = (RMI_RESULT_WORDS*16) - bit_delta;
    }
    return bit_delta;
}
/*
 * Make AST thread that continuously scans the system in the background.
 */
static void rmi_scan_post ( void *ctl_vp );	/* AST routine, forward reference */
static void rmi_next_sample ( struct rmi_async_control *ctl )
{
    int status;

    status = SYS$GETRMI ( ctl->scan_ef, 0, 0, rmi_item, &ctl->iosb, 
	rmi_scan_post, ctl );
    if ( (status&1) == 0 ) {
	ctl->abort = 1;
	rmi_scan_post ( ctl );
    }
}

static void rmi_scan_post ( void *ctl_vp )	/* AST routine */
{
    struct rmi_async_control *ctl;
    struct rmi_scan_result *cur;
    int status, bits_changed;

    ctl = ctl_vp;
    if ( ctl->abort ) {
	return;
    }
    if ( ctl->iosb.status == 0 ) {	/* Bugcheck, should have completed */
    } else if ( (ctl->iosb.status &1) == 0 ) {
	printf ( "Error returned by $GETRMI: %d\n", ctl->iosb.status );
	ctl->abort = 1;
    }
    /*
     * Make current buffer the previous buffer and fill new current.
     */
    cur = ctl->prev;
    ctl->prev = ctl->cur;
    ctl->cur = cur;
    status = condense_scan ( ctl->debug ? dbgfp : 0, cur );
    ctl->sample_num++;
    /*
     * Gather statistics for measuring randomness of result.
     */
    bits_changed = collect_quality ( cur, ctl->prev, &ctl->sq );
    /*
     * Schedule next sample to start after delay.
     */
    status = SYS$SETIMR ( ctl->scan_ef, &ctl->sample_interval, 
		rmi_next_sample, ctl, 0 );
    /*
     * Notify Non-AST level that next read completed.
     */
    if ( ctl->pending ) {
	if ( ctl->debug ) fprintf(dbgfp,
	    "transfering RMI data, quals: %d %d\n",
	    ctl->sq.lt_qual, ctl->sq.st_qual );
	*ctl->pending = *cur;
	ctl->pending = 0;
	ctl->pend_sample_num = ctl->sample_num;
	ctl->pend_quality = ctl->sq.st_qual;
	ctl->pend_lt_quality = ctl->sq.lt_qual;
	SYS$SETEF ( ctl->stall_ef );
    }

    return;
}

/*****************************************************************************/
/* Allocate a new rmi_async_control structure and initiate new thread.
 */
struct rmi_async_control  *rmi_async_init ( int interval_ms, int debug )
{
    struct rmi_async_control *ctl;
    int status;

    ctl = calloc ( 1, sizeof(struct rmi_async_control) );
    LIB$GET_EF ( &ctl->scan_ef );
    LIB$GET_EF ( &ctl->stall_ef );
    ctl->abort = 0;
    ctl->debug = debug;
    ctl->sq.debug = (debug > 1);
    ctl->cur = calloc ( 1, sizeof(struct rmi_scan_result) );
    ctl->prev = calloc ( 1, sizeof(struct rmi_scan_result) );

    
    ctl->sample_interval = interval_ms;
    ctl->sample_interval *= -10000;	/* convert MS to 100-nanosec ticks */

    status = SYS$DCLAST ( rmi_next_sample, ctl, 0 );
    if ( (status&1) == 0 ) ctl->abort = 1;

    return ctl;

}

/*****************************************************************************/
/* Main entry point called by random.c to get random data for seeding PRNG.
 *
 * Arguments:
 *    add:       function to transfer random bytes generated here
 *               to the consumer's buffer.
 *    requester: passed through to add() as "source".
 *    length:    bytes requested?
 *    level:     source code.  (See random.c: add_randomness().)
 */

#define MBX_BASE_NAME "MBX_GNUPG_ENTROPY_"

#ifndef TEST_RMI
int
rndvms_gather_random( void (*add)(const void*, size_t, int),
                      int requester, size_t length, int level )
{
    static struct rmi_async_control *ctl = 0;
    static struct rmi_scan_result cur;
    static int rmi_scan_avail = 0;
    static int rmi_scan_used = 0;
    /*
     * Start background data gathering if first call.
     */
    if ( !ctl ) {
	int interval, debug;
	char *evp;
        /* Open debug file, if specified. */
        dbgfn = getenv(DEBUG_ENV_VAR);
        if (dbgfn != NULL)
        {
            if (strcmp( dbgfn, "-") == 0)
            {
                dbgfp = stdout;
            }
            else
            {
                dbgfp = fopen( dbgfn, "w");
                if (dbgfp == NULL)
                {
                    g10_log_info(                     "rndvms()  Can't open (write) rnd debug file \"%s\".\n",
                     dbgfn);
                    g10_log_info( " %s\n", strerror( errno));
                }
            }
        }
	/*
	 * Create AST thread.
	 */
	evp = getenv ( INTERVAL_ENV_VAR );
	interval = (evp) ? atoi(evp) : RMI_DEFAULT_INTERVAL;
	debug = (dbgfp) ? 1 : 0;	/* default value */
	evp = getenv ( DEBUG_AST_ENV_VAR );
	if ( evp ) debug = atoi ( evp );

	ctl = rmi_async_init ( interval, debug );
	if ( !ctl ) return -1;
    }
/****************
 * Using a level of 0 should never block and better add nothing
 * to the pool.  So this is just a dummy for this gatherer.
 */
    if ( ctl->debug ) fprintf(dbgfp,"Request to gather %d L%d bytes...\n",
	length, level );
    if (level == 0)
        return 0;


    /* While more data remain to be supplied, supply them. */
    while (length > 0)
    {
	int buf_status, segment, eff_segment;
	char *bufp;

	if ( rmi_scan_avail <= rmi_scan_used ) {
	    buf_status = rmi_await_next ( ctl, &cur );
	    if ( !buf_status ) return -1;
	    rmi_scan_avail = RMI_RESULT_WORDS * 2;
	    rmi_scan_used = 0;
	}

	segment = rmi_scan_avail - rmi_scan_used;
	if ( segment > length ) segment = length;
	bufp = (char *) &cur;
	if ( level == 1 ) {
	    /* Low quality data is good enough */
            if (dbgfp)
            {
                fprintf( dbgfp,
                 "rndvms()  Adding %d \"random\" L1 bytes.\n", segment );
                fflush( dbgfp);
            }
	    (*add) ( &bufp[rmi_scan_used], segment, requester );
	    eff_segment = segment;
	    rmi_scan_used += segment;
	} else {
	    /*
	     * High quality data is oversampled.  Send entire contents to
	     * caller, but only count as (q*n)/256 effective bytes.
	     * Jitter the number of bytes sent by 0-3 bytes.
	     */
	    segment -= (ctl->pend_quality&3);
            if (dbgfp)
            {
                fprintf( dbgfp,
                 "rndvms()  Adding %d \"random\" L2 bytes(q=%d).\n", 
			segment, ctl->pend_quality );
                fflush( dbgfp);
            }
	    if ( segment > 0 ) {
		(*add) ( &bufp[rmi_scan_used], segment, requester );
		eff_segment = (segment*ctl->pend_quality) / (RMI_RESULT_WORDS*16);
		if ( eff_segment < 2 ) eff_segment = 2;
	    } else eff_segment = 0;
	    
	    rmi_scan_used = rmi_scan_avail;	/* buffer always fully consumed */
	}
	length -= eff_segment;
    }
    return 0;
}
#else
/*
 * Test program.  Command line: rndvms-rmi [interval] [debug]
 */
int main ( int argc, char **argv )
{
    int status, debug;
    int i, j, ind, interval;
    struct rmi_scan_result cur;
    struct rmi_async_control *ctl;
    /*
     * Poll each second.
     */
    interval = (argc>1) ? atoi(argv[1]) : RMI_DEFAULT_INTERVAL;
    debug = (argc > 2) ? atoi(argv[2]) : 0;
    dbgfp = stdout;
    ctl = rmi_async_init ( interval, debug );

    status = rmi_sample ( (argc>2) ? dbgfp : 0, &cur );
    while ( status & 1 ) {
	ind = rmi_await_next ( ctl, &cur );
        if ( ind == 0 ) status = 0;

	printf ( "Indicator:%3d, quality %3d %3d, - ",
	    ind, ctl->pend_lt_quality, ctl->pend_quality );
	for ( i = 0; i < RMI_RESULT_WORDS; i++ )
	    printf ( "%04x", cur.uw[i] );
	printf ( "\n" );
    }
    return status;
}
#endif
